"Code to process data and fit model parameters."
from __future__ import annotations

from multiprocessing import Pool, cpu_count
from typing import Callable

import numpy as np
import pandas as pd
import structlog
from scipy.special import binom
from textdistance import hamming
from tqdm import tqdm

log = structlog.get_logger(__name__)


def applyParallel(
    dfGrouped: list,
    func: Callable,
    cpuCount: int = cpu_count(),
    silent=False,
) -> pd.DataFrame:
    """Parallely runs func on each group of dfGrouped.

    Args:
        dfGrouped (list): Func runs parallely on each element of the list dfGrouped
        func (Callable): Function to run on dfGrouped
        cpuCount (int, optional): Number of cpus to use. Defaults cpu_count() the maximum possible.
        silent (bool): if true do not show progress bars.

    Returns:
        pd.Dataframe: Dataframe concatenating output of func on each group.
    """
    with Pool(cpuCount) as p:
        ret_list = list(
            tqdm(p.imap(func, dfGrouped), total=len(dfGrouped), disable=silent),
        )
    if len(ret_list) == 0:
        return pd.DataFrame()
    return pd.concat(ret_list)


def count_mutations(args: tuple[int, pd.DataFrame]):
    """Compute & return Return mutation counts column for a given dataframe.

    Args:
        args (tuple[int, pd.DataFrame]): _,dataframe of sequences

    Returns:
        pd.Dataframe: Dataframe with mutation counts.
    """
    _, df = args
    return df[["alt_sequence_alignment", "alt_germline_alignment"]].apply(
        lambda x: hamming(*x),
        axis=1,
    )


def preprocess(
    dataframe: pd.DataFrame,
    max_mutation_count: int = 60,
    lengths: np.ndarray = np.arange(15, 81 + 3, 3).astype(int),
    silent: bool = False,
) -> pd.DataFrame:
    """Processes input dataframe.
    Drop nulls, filters for cdr3 length & mutation counts.

    Args:
        dataframe (pd.DataFrame): Input dataframe of sequences.
        max_mutation_count (int, optional): Remove sequences with higher mutation count. \
            Defaults to 60.
        lengths (np.ndarray, optional): Remove sequences with CDR3 length not in lengths. \
            Defaults to np.arange(15, 81 + 3, 3).astype(int).
        silent (bool, optional): Do not show progress bar if true. Defaults to False.

    Returns:
        pd.Dataframe: processed dataframe.
    """
    df = dataframe.copy()[
        [
            "sequence_id",
            "v_call",
            "j_call",
            "junction",
            "v_sequence_alignment",
            "j_sequence_alignment",
            "v_germline_alignment",
            "j_germline_alignment",
        ]
    ]
    df = df.dropna()
    usecols = [
        "sequence_id",
        "v_gene",
        "j_gene",
        "cdr3_length",
        "cdr3",
        "alt_sequence_alignment",
        "alt_germline_alignment",
        "mutation_count",
    ]
    if "v_gene" not in df.columns:
        df[["v_gene", "_"]] = df["v_call"].str.split("*", expand=True)
    if "j_gene" not in df.columns:
        df[["j_gene", "_"]] = df["j_call"].str.split("*", expand=True)
    if "cdr3" not in df.columns:
        df["cdr3"] = df["junction"].str[3:-3]
    df["cdr3_length"] = df["cdr3"].str.len()

    df["alt_sequence_alignment"] = (
        df["v_sequence_alignment"] + df["j_sequence_alignment"]
    )
    df["alt_germline_alignment"] = (
        df["v_germline_alignment"] + df["j_germline_alignment"]
    )

    df["mutation_count"] = applyParallel(
        df.groupby(["v_gene", "j_gene", "cdr3_length"]),
        count_mutations,
        silent=silent,
    )
    log.debug(
        "Filtering sequences",
        criteria_one="CDR3 length not multiple of three.",
        criteria_two="CDR3 length not in [15,81].",
        criteria_three="More than 60 sequence mutations from germline.",
        criteria_four="With a null column value.",
    )
    return df.query(
        "cdr3_length in @lengths and mutation_count <= @max_mutation_count",
    )[usecols].astype({"mutation_count": int, "cdr3_length": int})


def create_classes(df: pd.DataFrame) -> pd.Dataframe:
    """Create VJl classes.

    Args:
        df (pd.DataFrame): Processed dataframe of sequences.

    Returns:
        pd.DataFrame: Dataframe with classes.
    """
    classes = (
        df.groupby(["v_gene", "j_gene", "cdr3_length"])
        .size()
        .to_frame("sequence_count")
    ).reset_index()
    classes["pair_count"] = (
        classes["sequence_count"].apply(lambda x: binom(x, 2)).astype(int)
    )
    l_classes = (
        classes.groupby("cdr3_length")[["sequence_count", "pair_count"]]
        .sum()
        .reset_index()
    )
    l_classes["v_gene"] = "None"
    l_classes["j_gene"] = "None"
    classes = pd.concat([classes, l_classes], ignore_index=True).sort_values(
        "sequence_count",
        ascending=False,
    )
    classes["class_id"] = range(1, len(classes) + 1)
    classes.reset_index(drop=True, inplace=True)
    return classes



from __future__ import annotations

import os
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.special import factorial


class EM:
    """Finds the mixture distribution
    that explains the statistics of distances"""

    def __init__(self, l, h, model=326713, howmany=10, positives="geometric"):
        self.l = int(l)
        self.h = h.astype(int)[: self.l + 1]
        self.b = np.arange(self.l + 1, dtype=int)
        self.model = model
        self.const_p0 = self.readNull()
        self.howmany = howmany
        self.positives = positives

    def readNull(self):
        """Read estimated null distribution"""
        dirname = Path(os.path.dirname(__file__))
        cdfs = pd.read_csv(
            dirname / Path(f"data/cdfs_{self.model}.csv"),
        )
        cdf = cdfs.loc[cdfs["l"] == self.l].values[0, 1 : self.l + 1]
        return np.diff(cdf, prepend=[0], append=[1])

    def discreteExpectation(self, theta):
        """Calculate membership probabilities"""
        rho, mu = theta
        if self.positives == "geometric":
            P1 = rho / (mu + 1) * (mu / (mu + 1)) ** self.b
        elif self.positives == "poisson":
            P1 = rho * mu**self.b * np.exp(-mu) / factorial(self.b)
        P0 = (1 - rho) * self.const_p0[self.b]
        return np.array([P1, P0]) / (P1 + P0)

    def dicreteMaximization(self, theta):
        """Maximize current likelihood"""
        P1, P0 = self.discreteExpectation(theta)
        P1Sum, P0Sum = (self.h * P1).sum(), (self.h * P0).sum()
        rho = P1Sum / (P1Sum + P0Sum)
        mu = np.dot(self.h * P1, self.b) / P1Sum
        return rho, mu

    def discreteEM(self):
        """EM with discrete model:
        P1 geometric or Poisson with mean mu
        P0 estimated with Ppost"""
        mu = 0.02 * self.l
        rho = self.h[0] / sum(self.h) * (1 + mu)
        theta = (rho, mu)
        for _ in range(self.howmany):
            theta = self.dicreteMaximization(theta)
        return theta

    def discreteMix(self, theta):
        """Evaluate mixture distribution"""
        rho, mu = theta
        if self.positives == "geometric":
            return (
                rho / (mu + 1) * (mu / (mu + 1)) ** self.b
                + (1 - rho) * self.const_p0[self.b]
            )
        if self.positives == "poisson":
            return (
                rho * mu**self.b * np.exp(-mu) / factorial(self.b)
                + (1 - rho) * self.const_p0[self.b]
            )

    def error(self, theta, threshold=0.15, error="rrmse"):
        """Estimate goodness of fit
        By default returns rescaled root MSE"""
        y1 = self.h / sum(self.h)
        y2 = self.discreteMix(theta)
        mask = (y1 > 0) * (self.b < threshold * self.l)
        y1m, y2m = y1[mask], y2[mask]
        logy1, logy2 = np.log(y1m), np.log(y2m)

        mse = ((y1m - y2m) ** 2).sum() / mask.sum()
        rmse = np.sqrt(mse)
        msle = ((logy1 - logy2) ** 2).sum() / mask.sum()
        rmsle = np.sqrt(msle)
        mae = np.abs(y1m - y2m).sum() / mask.sum()
        dkl = (y2m * (logy2 - logy1)).sum() / np.log(2)
        if error == "rmse":
            return rmse
        elif error == "rmsle":
            return rmsle
        elif error == "mae":
            return mae
        elif error == "dkl":
            return dkl
        elif error == "rrmse":
            return rmse / theta[0]
        else:
            return False



from __future__ import annotations

import json
import logging
from multiprocessing import cpu_count
from pathlib import Path

import pandas as pd
import structlog
import typer

from hilary.apriori import Apriori
from hilary.inference import HILARy

app = typer.Typer(add_completion=False)


@app.command()
def main(
    data_path: Path = typer.Argument(
        ...,
        help="Path of the excel file to infer lineages.",
        show_default=False,
    ),
    verbose: int = typer.Option(
        0,
        "--verbose",
        "-v",
        count=True,
        help="Set logging verbosity level.",
    ),
    threads: int = typer.Option(
        cpu_count(),
        "--threads",
        "-t",
        help="Choose number of cpus on which to run code.",
    ),
    precision: float = typer.Option(
        0.99,
        "--precision",
        "-p",
        help="Choose desired precision.",
    ),
    sensitivity: float = typer.Option(
        0.9,
        "--sensitivity",
        "-s",
        help="Choose desired sensitivity.",
    ),
    nmin: int = typer.Option(
        1,
        "--nmin",
        help="Infer prevalence and mu on classes of size larger than nmin. \
            Mean prevalence is assigned to lower than nmin classes.",
    ),
    model: int = typer.Option(
        326713,
        "--model",
        "-m",
        help="Model name to infer Null distribution.",
    ),
    silent: bool = typer.Option(
        False,
        "--silent",
        help="Do not show progress bars if used.",
    ),
    result_folder: Path = typer.Option(
        None,
        "--result-folder",
        help="""Where to save the result files. By default it will be saved in a 'result/'
            folder in input data's parent directory.""",
    ),
    config: Path = typer.Option(
        None,
        "--config",
        help="""Configuration file for column names. File should be a json with keys as your data's
            column names and values as hilary's required column names.""",
    ),
) -> None:
    """Infer lineages from data_path excel file.

    Args:
        data_path (Path): Path of the excel file to infer lineages.
        verbose (int, optional): Set logging verbosity level, defaults to 0.
        threads (int, optional): Number of cpus on which to run code, defaults to cpu_count().
        precision (float, optional): Desired precision, defaults to 0.99.
        sensitivity (float, optional): Desired sensitivity, defaults to 0.9.
        nmin (int, optional): Infer prevalence and mu on classes of size > nmin, defaults to 100000.
        model (int, optional): Model name to infer Null distribution, defaults to 326713.
        silent (bool,optional): Do not show progress bars if used.
        result_folder (Path): Where to save the result files. By default it will be saved in a \
            'result/' folder in input data's parent directory.
        config (Path): Configuration file for column names. File should be a json with keys \
            as your data's column names and values as hilary's required column names.
    """
    if result_folder is None:
        result_folder = data_path.parents[0] / Path("hilary_results/")
    result_folder.mkdir(parents=True, exist_ok=True)
    debug_folder = result_folder / Path("debug/")
    debug_folder.mkdir(parents=True, exist_ok=True)

    if verbose >= 2:  # noqa: PLR2004
        logging_level = logging.DEBUG
    elif verbose == 1:
        logging_level = logging.INFO
    else:
        logging_level = logging.WARNING
    renderer = structlog.dev.ConsoleRenderer(sort_keys=False)
    structlog.configure(
        wrapper_class=structlog.make_filtering_bound_logger(logging_level),
        processors=[
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.processors.StackInfoRenderer(),
            renderer,
        ],
    )
    log = structlog.get_logger()
    log.info(
        "📖 READING DATA 📖.",
        data_path=data_path.as_posix(),
    )
    suffix = data_path.suffix
    if suffix == ".xlsx":
        dataframe = pd.read_excel(data_path)
    elif suffix == ".tsv":
        dataframe = pd.read_csv(
            data_path,
            sep="\t",
        )
    else:
        raise ValueError(
            f"Format {suffix} not supported. Extensions supported are tsv, xlsx.",
        )
    if config:
        with open(config, encoding="utf-8") as user_file:
            column_dict = json.load(user_file)
            for key in column_dict:
                dataframe[column_dict[key]] = dataframe[key]

    log.debug("Displaying dataframe columns.", columns=dataframe.columns)
    if logging_level == logging.DEBUG:
        input_path = debug_folder / Path(f"input_{data_path.name}").with_suffix(".xlsx")
        log.debug(
            "Saving input dataframe.",
            path=input_path.as_posix(),
        )
        dataframe.to_excel(input_path.as_posix())
    apriori = Apriori(
        dataframe,
        threads=threads,
        precision=precision,
        sensitivity=sensitivity,
        nmin=nmin,
        model=model,
        silent=silent,
    )
    if logging_level == logging.DEBUG:
        preprocessed_path = debug_folder / Path(
            f"preprocessed_input_{data_path.name}",
        ).with_suffix(".xlsx")
        log.debug(
            "Saving dataframe after preprocessing.",
            path=preprocessed_path.as_posix(),
        )
        classes_path = debug_folder / Path(
            f"classes_{data_path.name}",
        ).with_suffix(".xlsx")
        log.debug(
            "Saving classes after preprocessing.",
            path=classes_path.as_posix(),
        )
        apriori.df.to_excel(preprocessed_path.as_posix())
        apriori.classes.to_excel(classes_path.as_posix())
    log.info("⏳ COMPUTING HISTOGRAMS ⏳.")
    apriori.get_histograms()
    log.info("⏳ COMPUTING PARAMETERS ⏳.")
    apriori.get_parameters()
    log.info("⏳ COMPUTING THRESHOLDS ⏳.")
    apriori.get_thresholds()
    hilary = HILARy(apriori)
    log.info("⏳ COMPUTING PRECISE AND SENSITIVE CLUSTERS ⏳.")
    hilary.compute_prec_sens_clusters()
    log.info("⏳ INFERRING FAMILIES ⏳.")
    hilary.infer()
    if logging_level == logging.DEBUG:
        output_path = debug_folder / Path(
            f"preprocessed_output_{data_path.name}",
        ).with_suffix(".xlsx")
        log.debug(
            "Saving dataframe inferred by Hilary.",
            path=output_path.as_posix(),
        )
        hilary.df.to_excel(output_path)
        parameters_path = debug_folder / Path(
            f"parameters_{data_path.name}",
        ).with_suffix(".xlsx")
        log.debug(
            "Saving all parameters inferred by Hilary.",
            path=parameters_path.as_posix(),
        )
        hilary.classes.to_excel(parameters_path)
    mask = ~dataframe.index.isin(hilary.df.index)
    dataframe["family"] = hilary.df["family"]
    dataframe.loc[mask, "family"] = 0
    output_path = result_folder / Path(f"inferred_{data_path.name}").with_suffix(
        ".xlsx",
    )
    log.info("💾 SAVING RESULTS 💾.", output_path=output_path.as_posix())
    dataframe.to_excel(output_path)


if __name__ == "__main__":
    app()



from __future__ import annotations

from itertools import combinations
from multiprocessing import Pool, cpu_count

import numpy as np
import pandas as pd
import structlog
from atriegc import TrieNucl as Trie
from scipy.cluster.hierarchy import fcluster, linkage
from scipy.spatial.distance import squareform
from textdistance import hamming

from hilary.apriori import Apriori
from hilary.utils import applyParallel

# from atriegc import Trie

log = structlog.get_logger()


class CDR3Clustering:
    """Infer families using CDR3 length and thresholds computed by Apriori class."""

    def __init__(self, thresholds: pd.DataFrame) -> None:
        """Initialize thresholds.

        Args:
            thresholds pd.DataFrame: Dataframe containing thresholds for each (V,J,l) class.
        """
        self.thresholds = thresholds

    def cluster(self, args: tuple[tuple[str, str, int], pd.DataFrame]) -> pd.Series:
        """Returns cluster labels depending of thresholds in self.thresholds

        Args:
            args (Tuple[Tuple[str, str, int], pd.DataFrame]): (Vgene,Jgene,l), Dataframe
            of sequences grouped by V,J,l class.

        Returns:
            pd.Series: Cluster labels for this V,J,l class.
        """
        (v, j, l), df = args
        trie = Trie()
        for cdr3 in df["cdr3"]:
            trie.insert(cdr3)
        t = self.thresholds.loc[
            (self.thresholds.v_gene == v)
            & (self.thresholds.j_gene == j)
            & (self.thresholds.cdr3_length == l)
        ].values[0][-1]
        if t >= 0:  # ? question t>0 better ?
            dct = trie.clusters(t)
            return df["cdr3"].map(dct)
        return df.index.to_frame()

    def infer(
        self,
        df: pd.DataFrame,
        group: list[str] = ["v_gene", "j_gene", "cdr3_length"],
        silent: bool = False,
    ) -> pd.Series:
        """Returns cluster labels depending of thresholds in self.thresholds.
        Runs self.cluster parallely on dataframe grouped by 'group' argument.

        Args:
            df (pd.DataFrame): Dataframe of sequences.
            group (list[str], optional): Groups on which to do parallel inferring of clusters.
            Defaults to ["v_gene", "j_gene", "cdr3_length"].
            silent (bool,optional) : Do not show progress bar if True.

        Returns:
            pd.Series: Series with cluster labels.
        """
        use = group + ["cdr3"]
        df["cluster"] = applyParallel(
            df[use].groupby(group),
            self.cluster,
            silent=silent,
        )
        group = group + ["cluster"]
        return df.groupby(group).ngroup() + 1


class DistanceMatrix:
    """Object enabling parallel computing of the distance martrix of a large cluster."""

    def __init__(
        self,
        l: int,
        df: pd.DataFrame,
        threads: int = cpu_count() - 1,
    ) -> None:
        """Initialize attributes.

        Args:
            l (int): CDR3 length
            df (pd.DataFrame): Dataframe of sequences grouped by (v,j,l,sensitive cluster)
            threads (int, optional): Number of cpus on which to run code. Defaults to cpu_count()-1.
        """
        self.threads = threads
        self.l = l
        self.L = 250
        self.l_L = l / self.L
        self.l_L_L = l / (l + self.L)

        self.data = df.values
        self.n = self.data.shape[0]
        # maximum elements in 1D dist array
        self.k_max = self.n * (self.n - 1) // 2
        self.k_step = self.n**2 // 2 // (500)  # ~500 bulks

    def metric(
        self,
        arg1: tuple[int, str, int, int],
        arg2: tuple[int, str, int, int],
    ) -> float:
        """Compute difference btween normalized cdr3 divergence & shared mutations of two sequences.

        Args:
            arg1 (Tuple[int,str,int,int]): (CDR3 length, V+J sequence alignment, number of mutations
            from germline, index) for sequence 1
            arg2 (Tuple[int,str,int,int]): Same for sequence 2

        Returns:
            float: Difference between two quantities.
        """
        cdr31, s1, n1, i1 = arg1
        cdr32, s2, n2, i2 = arg2
        if i1 == i2:
            return -100.0
        if n1 * n2 == 0:
            return 100.0
        n = hamming(cdr31, cdr32)
        nL = hamming(s1, s2)
        n0 = (n1 + n2 - nL) / 2

        exp_n = self.l_L * (nL + 1)
        std_n = np.sqrt(exp_n * self.l_L_L)

        exp_n0 = n1 * n2 / self.L
        std_n0 = np.sqrt(exp_n0)

        x = (n - exp_n) / std_n
        y = (n0 - exp_n0) / std_n0
        return x - y

    def proc(self, start: int) -> tuple[int, int, list[float]]:
        """Compute 1D distance matrix between start and start+self.k_step

        Args:
            start (int): Index from which to compute distances.

        Returns:
            Tuple[int, int, list[float]]: Start index, End index, distance for indices inbetween
        """
        dist = []
        k1 = start
        k2 = min(start + self.k_step, self.k_max)
        for k in range(k1, k2):
            # get (i, j) for 2D distance matrix knowing (k) for 1D distance matrix
            i = int(
                self.n
                - 2
                - int(np.sqrt(-8 * k + 4 * self.n * (self.n - 1) - 7) / 2.0 - 0.5),
            )
            j = int(
                k
                + i
                + 1
                - self.n * (self.n - 1) / 2
                + (self.n - i) * ((self.n - i) - 1) / 2,
            )
            # store distance
            a = self.data[i, :]
            b = self.data[j, :]
            d = self.metric(a, b)
            dist.append(d)
        return k1, k2, dist

    def compute(self) -> np.ndarray:
        """Run self.proc parallely to compute 1D distance matrix.

        Returns:
            np.array: 1D distance matrix
        """
        dist = np.zeros(self.k_max)
        with Pool(self.threads) as pool:
            for k1, k2, res in pool.imap_unordered(
                self.proc,
                range(0, self.k_max, self.k_step),
            ):
                dist[k1:k2] = res
        return dist + 100.0


class HILARy:
    """Infer families using CDR3 and mutation information"""

    def __init__(self, apriori: Apriori):
        """Initialize Hilary attributes using Apriori object.

        Args:
            apriori (Apriori): Apriori object containing histograms and thresholds.
        """

        self.group = ["v_gene", "j_gene", "cdr3_length"]
        self.df = apriori.df
        self.classes = apriori.classes
        self.use = [
            "cdr3",
            "alt_sequence_alignment",
            "mutation_count",
            "index",
        ]
        self.alignment_length = 250
        self.x0 = 0.0
        self.y0 = 0.0
        self.remaining = (
            self.classes.query(
                "v_gene != 'None' and precise_threshold < sensitive_threshold and pair_count > 0",
            )
            .groupby(self.group)
            .first()
            .index
        )
        self.silent = apriori.silent

    def singleLinkage(
        self,
        indices: np.ndarray,
        dist: np.ndarray,
        threshold: float,
    ) -> dict[int, int]:
        """Maps precise clusters to new precise AND sensitive clusters by merging clusters together.


        Args:
            indices (np.array): Indices of precise clusters.
            dist (np.ndarray): Distances between precise clusters.
            threshold (float): Threshold to merge two precise clusters if the distance is smaller.

        Returns:
            dict: _description_
        """
        clusters = fcluster(
            linkage(dist, method="single"),
            criterion="distance",
            t=threshold,
        )
        return {i: c for i, c in zip(indices, clusters)}

    def class2pairs(
        self,
        args: tuple[tuple[str, str, int, int], pd.DataFrame],
    ) -> pd.Series:
        """Group precise clusters not reaching desired sensitivity together.

        Args:
            args (Tuple[Tuple[str,str,int,int],pd.DataFrame]): (Vgene,Jgene,cdr3length),dataframe
            representing sensitive cluster.

        Returns:
            pd.Series: Precise clusters which now reach required sensitivity.
        """
        (
            (
                _,
                _,
                l,
                _,
            ),
            df,
        ) = args  # (vgene, jgene, cdr3length, sensitive cluster), df
        indices = np.unique(df["precise_cluster"])
        if len(indices) <= 1:
            return df["precise_cluster"]

        translateIndices = dict(zip(indices, range(len(indices))))
        df["index"] = df["precise_cluster"].map(translateIndices)
        dim = len(indices)
        distanceMatrix = np.ones((dim, dim), dtype=float) * (200)
        for i in range(dim):
            distanceMatrix[i, i] = 0

        for (cdr31, s1, n1, i1), (cdr32, s2, n2, i2) in combinations(
            df[self.use].values,
            2,
        ):
            if i1 == i2 or n1 * n2 == 0:
                continue
            n = hamming(cdr31, cdr32)
            nL = hamming(s1, s2)
            n0 = (n1 + n2 - nL) / 2

            exp_n = l / self.alignment_length * (nL + 1)
            std_n = np.sqrt(
                exp_n * (l + self.alignment_length) / self.alignment_length,
            )

            exp_n0 = n1 * n2 / self.alignment_length
            std_n0 = np.sqrt(exp_n0)

            x = (n - exp_n) / std_n
            y = (n0 - exp_n0) / std_n0
            distance = x - y + 100.0

            distanceMatrix[i1, i2] = distance
            distanceMatrix[i2, i1] = distance

        sl = self.singleLinkage(
            indices,
            squareform(distanceMatrix),
            threshold=self.x0 - self.y0 + 100.0,
        )
        return df["precise_cluster"].map(sl)

    def compute_prec_sens_clusters(self) -> None:
        """Infer precise and sensitive clusters."""
        prec = CDR3Clustering(self.classes[self.group + ["precise_threshold"]])
        sens = CDR3Clustering(
            self.classes[self.group + ["sensitive_threshold"]],
        )
        self.df["precise_cluster"] = prec.infer(self.df, silent=self.silent)
        self.df["sensitive_cluster"] = sens.infer(self.df, silent=self.silent)

    def mark_class(self, df: pd.DataFrame) -> pd.Series:
        """Flag all indices of a sensitive cluster not reaching desired sensitivity.

        Args:
            df (pd.DataFrame): Dataframe grouped representing given cluster.

        Returns:
            pd.Series: Series flagging all indices of given cluster.
        """
        df["to_resolve"] = True
        return df["to_resolve"]

    def to_do(self, size_threshold: float = 1e3) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Classify sensitive clusters not reaching desired sensitivity into big or small cluster.

        Args:
            size_threshold (float, optional): Threshold to separate big and small clusters.
            Defaults to 1e3.

        Returns:
            Tuple[pd.DataFrame,pd.DataFrame]: Returns indices of small and big sensitive clusters.
        """
        self.df["to_resolve"] = False
        self.df["to_resolve"] = applyParallel(
            [self.df.groupby(self.group).get_group(g) for g in self.remaining],
            self.mark_class,
            silent=True,
        )
        self.df.fillna(value={"to_resolve": False}, inplace=True)
        dfGrouped = self.df.query("to_resolve == True").groupby(
            self.group + ["sensitive_cluster"],
        )
        sizes = dfGrouped.size()
        mask = sizes > size_threshold
        large_to_do = sizes[mask].index
        small_to_do = sizes[~mask].index
        return small_to_do, large_to_do

    def infer(self) -> None:
        """Infer family clusters.
        First, for each sensitive cluster that does not reach desired sensitivity, group precise
        clusters together with a single linkage algorithm. This grouping is done differently
        depending on whether the sensitive cluster is large or not.
        """
        if self.remaining.empty:
            self.df["family"] = self.df["precise_cluster"]
            self.df = self.df.drop(
                columns=[
                    "cluster",
                    "precise_cluster",
                    "sensitive_cluster",
                ],
            )
            return
        small_to_do, large_to_do = self.to_do()
        dfGrouped = self.df.groupby(self.group + ["sensitive_cluster"])
        log.debug(
            "Grouping precise clusters together to reach desired sensitivity.",
        )
        self.df["family_cluster"] = applyParallel(
            [(g, dfGrouped.get_group(g)) for g in small_to_do],
            self.class2pairs,
            silent=self.silent,
        )
        for g in large_to_do:
            _, _, l, _ = g  # (vgene, jgene, cdr3length, sensitive cluster)
            dm = DistanceMatrix(
                l,
                dfGrouped.get_group(g)[
                    [
                        "cdr3",
                        "alt_sequence_alignment",
                        "mutation_count",
                        "precise_cluster",
                    ]
                ],
            )

            d = dm.compute()
            dct = self.singleLinkage(
                dfGrouped.get_group(g).index,
                d,
                self.x0 - self.y0 + 100.0,
            )
            self.df["family_cluster"] = self.df.index.map(dct)

        self.df.fillna(value={"family_cluster": 0}, inplace=True)
        self.df["family"] = (
            self.df.groupby(
                self.group + ["sensitive_cluster", "family_cluster"],
            ).ngroup()
            + 1
        )
        self.df = self.df.drop(
            columns=[
                "family_cluster",
                "cluster",
                "precise_cluster",
                "sensitive_cluster",
                "to_resolve",
            ],
        )
